
#property library
#include <stderror.mqh>


/**
 * Reads entire file into an array. Each element of the array corresponds to a line in the file,
 * with the newline separator removed.  Returns the number of resulting elements/lines.
 *
 * @param  string filename       - filename relative to .\experts\files
 * @param  string result[]       - array for resulting lines
 * @param  bool   skipEmptyLines - whether or not to skip empty lines
 *
 * @return int - number of resulting lines or -1 in case of any errors
 */
int FileReadLines(string filename, string result[], bool skipEmptyLines) {
   int hFile, hFileBin, fieldSeparator='\t';

   // open file
   hFile = FileOpen(filename, FILE_CSV|FILE_READ, fieldSeparator);         // filename relative to .\experts\files
   if (hFile < 0) {
      catch("FileReadLines(1)->FileOpen(\""+ filename +"\")");
      return(-1);
   }


   // quick return for empty files
   if (FileSize(hFile) == 0) {
      FileClose(hFile);
      ArrayResize(result, 0);
      return(0);
   }


   // read file line by line
   bool newLine=true, blankLine=false, lineEnd=true, wasSeparator;
   string line, value, lines[]; ArrayResize(lines, 0);                     // lines[]: tmp. storage for readed lines
   int i, len, fPointer;                                                   // line counter and length of read string

   while (!FileIsEnding(hFile)) {
      newLine = false;
      if (lineEnd) {                                                       // If the last loop encountered end of line
         newLine   = true;                                                 // we reset all flags to start of new line.
         blankLine = false;
         lineEnd   = false;
         fPointer  = FileTell(hFile);                                      // always points to start of current line
      }

      // read line
      value = FileReadString(hFile);

      // check for line or file end (whatever comes first)
      if (FileIsLineEnding(hFile) || FileIsEnding(hFile)) {
         lineEnd  = true;
         if (newLine) {
            if (StringLen(value) == 0) {
               if (FileIsEnding(hFile))                                    // start_of_line + empty + end_of_file => in fact nothing -> break
                  break;
               blankLine = true;                                           // start_of_line + empty + end_of_line => blank line
            }
         }
      }

      // skip blank lines
      if (blankLine) /*&&*/ if (skipEmptyLines)
         continue;

      // store read value in new line or update previous line
      if (newLine) {
         i++;
         ArrayResize(lines, i);
         lines[i-1] = value;
      }
      else {
         // FileReadString() reads max. 4095 chars: check last char of long lines for separator
         len = StringLen(lines[i-1]);
         if (len < 4095) {
            wasSeparator = true;
         }
         else {
            // long line: re-open the same file in binary mode and check the character where FileReadString() stopped for separator
            if (hFileBin == 0) {
               hFileBin = FileOpen(filename, FILE_BIN|FILE_READ);
               if (hFileBin < 0) {
                  FileClose(hFile);
                  catch("FileReadLines(3)->FileOpen(\""+ filename +"\")");
                  return(-1);
               }
            }
            if (!FileSeek(hFileBin, fPointer+len, SEEK_SET)) {
               FileClose(hFile);
               FileClose(hFileBin);
               catch("FileReadLines(4)->FileSeek(hFileBin, "+ (fPointer+len) +", SEEK_SET)", GetLastError());
               return(-1);
            }
            wasSeparator = (fieldSeparator == FileReadInteger(hFileBin, CHAR_VALUE));
         }

         if (wasSeparator) lines[i-1] = StringConcatenate(lines[i-1], CharToStr(fieldSeparator), value);
         else              lines[i-1] = StringConcatenate(lines[i-1],                            value);
      }
   }

   // end of file triggered setting of ERR_END_OF_FILE
   int error = GetLastError();
   if (error!=ERR_END_OF_FILE) /*&&*/ if (IsError(error)) {
      FileClose(hFile);
      if (hFileBin != 0)
         FileClose(hFileBin);
      catch("FileReadLines(5)", error);
      return(-1);
   }

   // close open files
   FileClose(hFile);
   if (hFileBin != 0)
      FileClose(hFileBin);

   // copy lines from tmp. storage to result[]
   ArrayResize(result, i);
   if (i > 0)
      ArrayCopy(result, lines);

   if (ArraySize(lines) > 0)
      ArrayResize(lines, 0);

   if (IsError(catch("FileReadLines(6)")))
      return(-1);
   return(i);
}


// -----------------------------------------------------------------------------------------------------------------
// error handling, little helpers etc., unrelated to FileReadLines()
// -----------------------------------------------------------------------------------------------------------------


#import "user32.dll"
   int  MessageBoxA(int hWnd, string lpText, string lpCaption, int style);
#import
#define MB_OK              0x00000000
#define MB_ICONERROR       0x00000010


#import "winmm.dll"
   bool PlaySoundA(string lpSound, int hMod, int fSound);
#import
#define SND_ASYNC                0x01     // play asynchronously
#define SND_FILENAME       0x00020000     // name is file name


/**
 * Error checking/handling
 *
 * @param  string location - location identifier of the error occured
 * @param  int    error    - enforce a specific error by hand (default: ERR_NO_ERROR)
 *
 * @return int - the error code itself
 */
int catch(string location, int error=ERR_NO_ERROR) {
   if (error == ERR_NO_ERROR)
      error = GetLastError();

   if (error != ERR_NO_ERROR) {
      string message = location +"  [error code: "+ error +"]";

      // log error
      Alert("ERROR:   ", Symbol(), ",  ", WindowExpertName(), "::", message);

      // show error
      if (IsTesting()) {
         // force a sound and a message box even if MT4 wouldn't do so (Alert()/MessageBox() are not available in Tester)
         PlaySoundA(TerminalPath() +"\\sounds\\alert.wav", NULL, SND_FILENAME|SND_ASYNC);
         MessageBoxA(NULL, "ERROR in "+ WindowExpertName() +"::"+ message, "Strategy Tester", MB_ICONERROR|MB_OK);
      }
   }
   return(error);
}


/**
 * Quick check if given value represents an error, for improved readability.
 *
 * @param  int value
 *
 * @return bool
 */
bool IsError(int value) {
   return(value != ERR_NO_ERROR);
}
